/*
 *  msutil.c
 *
 *  Contained here is a miscellaneous collection of routines written
 *  specifically for use with version 4.0 of the Microsoft C Compiler.
 *  The purpose of the functions ranges from string manipulation to
 *  windowing to terminal i/o.  Each function carries with it a brief
 *  explanation of its operation, its parameters and its return value.
 *
 *  I'd sincerely appreciate any sort of feedback (good or bad) on the
 *  usefulness of these utilities, and suggestions for any enhancements
 *  or additions to the set.  Comments can be directed to me via the
 *  bulletin board or to:
 *
 *                           Scott Vaillancourt
 *                           1528 Asbury Street
 *                           St. Paul, MN  55108
 *                           Phone:  612/646-3159
 *
 *  Copyright (C) Scott Vaillancourt, 1987
 */

#include <conio.h>
#include <ctype.h>
#include <dos.h>
#include <io.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "msutil.h"

movch (sstr, sch, dstr, dch, numch)
/*
 *  This routine moves one or more characters from a given location in a
 *  source string to a given location in a destination string.
 *  The caller should take care that there is enough space in the
 *  destination string to accomodate the characters to be moved.
 *
 *  Parameters:
 *     sstr:   pointer to the source string
 *     sch:    starting character in the source string (zero based)
 *     dstr:   pointer to the destination string
 *     dch:    starting character in the destination string (zero based)
 *     numch:  number of characters to move
 *
 *  Return value:
 *     none
 */
int  sch, dch, numch;
char *sstr, *dstr;
   {
   sstr+=sch;                          /* adjust the source pointer */
   dstr+=dch;                          /* adjust the destination pointer */
   strncpy(dstr, sstr, numch);         /* move the characters */
   }

inkey ()
/*
 *  Waits for a keystroke and returns either the ascii code for the
 *  character or, if an extended character sequence was entered, the scan
 *  code of the key plus 256.  (For a list of the standard and extended
 *  characters, refer to Appendix G in the IBM BASIC Manual.)
 *
 *  Parameters:
 *     none
 *
 *  Return value:
 *     the ascii value or scan code of the key that was pressed
 */
   {
   int scanbyte;

   scanbyte=getch();
   if (scanbyte == 0)                  /* check for extended sequence */
      {
      scanbyte=getch();
      scanbyte+=256;                   /* set special character */
      }
   return(scanbyte);
   }

strblnk (str, len)
/*
 *  Set all characters in a string to blank and affix a null terminator
 *  to the end of the string.
 *
 *  Parameters:
 *     str:  string to be initialized
 *     len:  total number of characters in the string (including '\0')
 *
 *  Return value:
 *     none
 */
char str[];
int  len;
   {
   memset(str, 0x20, --len);           /* blank the string */
   str[len]='\0';                      /* now terminate it */
   }

trimb (str)
/*
 *  This routine deletes trailing blanks from a character string.  When the
 *  trimming has taken place, the length of the altered string is returned.
 *
 *  Parameters:
 *     str:  the character string to be trimmed
 *
 *  Return value:
 *     the length of the string after trimming has taken place
 */
register char *str;
   {
   register char *ptr;

   ptr=str + strlen(str) - 1;          /* point to last char in line */
   while (*ptr==' ' && ptr>=str)
      *ptr-- = '\0';                   /* trim another blank */
   return (strlen(str));               /* return the new length */
   }

trimall (str)
/*
 *  This routine deletes whitespace characters from a string.  When the
 *  trimming has taken place, the length of the altered string is returned.
 *
 *  Parameters:
 *     str:  the character string to be trimmed
 *
 *  Return value:
 *     the length of the string after trimming has taken place
 */
char *str;
   {
   register char *ptr;

   ptr=str + strlen(str) - 1;          /* point to last char in line */
   while (isspace(*ptr) && ptr>=str)
      *ptr-- = '\0';                   /* lop off another character */
   return (strlen(str));               /* return the new length */
   }

inschar (newch, str, loc)
/*
 *  Insert a character into a string at a user-specified location and
 *  adjust the positions of all succeeding characters.
 *
 *  Parameters:
 *     newch:  character to be inserted.
 *     str:    string into which the character should be inserted.
 *     loc:    location in 'str' where character should be inserted.
 *
 *  Return value:
 *     none
 */
char newch, *str;
int  loc;
   {
   register char *ptr1, *ptr2;

   ptr1=str + strlen(str);
   ptr2=ptr1 + 1;
   str+=loc;
   while (ptr1 >= str)                 /* adjust the characters up */
      *ptr2-- = *ptr1--;
   *str=newch;                         /* lay in the new character */
   }

delchar (str, loc)
/*
 *  Delete a character in a string at a user-specified location and
 *  adjust the positions of all succeeding characters.
 *
 *  Parameters:
 *     str:    string into which the character should be deleted.
 *     loc:    location in 'str' where character should be deleted.
 *
 *  Return value:
 *     none
 */
char *str;
int  loc;
   {
   register char *ptr;

   str+=loc;
   ptr=str+1;
   while (*str)
      *str++ = *ptr++;
   }

ismono ()
/*
 *  Determine the type of video adapter is being used.
 *
 *  Parameters:
 *     none
 *
 *  Return value:
 *     TRUE if a monochrome adapter is being used and FALSE if not
 */
   {
   union REGS inregs, outregs;

   int86(0x11, &inregs, &outregs);     /* get the equipment list */
   outregs.x.ax &= 0x0030;             /* mask out the video bits */
   return (!(outregs.x.ax ^ 0x0030));
   }

curset (onflag, first, last)
/*
 *  This routine is used to control the size and appearance of the cursor.
 *  The cursor may be turned either on or off, or its size may be adjusted.
 *
 *  Parameters:
 *     onflag:  flag indicating cursor status--TRUE will cause the cursor
 *              to be displayed and FALSE will turn it off
 *     first:   starting scan line for cursor
 *     last:    ending scan line for cursor
 *
 *  Return value:
 *     none
 */
char onflag;
int  first, last;
   {
   union REGS inregs, outregs;

   inregs.h.ah=1;                      /* service code to set cursor type */
   inregs.h.ch=(char)first;            /* first line of cursor */
   inregs.h.cl=(char)last;             /* last line of cursor */
   if (onflag)
      inregs.h.ch &= 0x1f;             /* turn cursor on (bit 5=0) */
   else
      inregs.h.ch |= 0x20;             /* turn cursor off (bit 5=1) */
   int86(0x10, &inregs, &outregs);     /* execute video interrupt */
   }

locate (row, col)
/*
 *  Position cursor on the screen.
 *
 *  Parameters:
 *     row:  row in which cursor is to be placed (1-25)
 *     col:  column in which cursor is to be placed (1-80)
 *
 *  Return value:
 *     none
 */
int  row, col;
   {
   union REGS inregs, outregs;

   inregs.h.ah=2;                      /* service code to position cursor */
   inregs.h.dh=--row;                  /* cursor row */
   inregs.h.dl=--col;                  /* cursor column */
   inregs.h.bh=0;                      /* page number */
   int86(0x10, &inregs, &outregs);     /* execute video interrupt */
   }

upscroll (ulr, ulc, lrr, lrc, lines)
/*
 *  Scroll a window on the screen up a given number of lines.  If the
 *  number of lines to be scrolled exceeds the number of rows available
 *  in the window, then the area is cleared rather than scrolled.
 *
 *  Parameters:
 *     ulr:    upper, left row of window
 *     ulc:    upper, left column of window
 *     lrr:    lower, right row of window
 *     lrc:    lower, right column of window
 *     lines:  number of lines to scroll window
 *
 *  Return value:
 *     none
 */
int ulr, ulc, lrr, lrc, lines;
   {
   union REGS inregs, outregs;

   inregs.h.ah=0x06;                   /* service code to scroll up */
   inregs.h.bh=0x07;                   /* attribute for new lines */
   inregs.h.ch=--ulr;                  /* upper left row */
   inregs.h.cl=--ulc;                  /* upper left column */
   inregs.h.dh=--lrr;                  /* lower right row */
   inregs.h.dl=--lrc;                  /* lower right column */
   if ((lrr-ulr+1) > lines)            /* see if window is big enough */
      inregs.h.al=lines;               /* number of lines to scroll */
   else
      inregs.h.al=0;                   /* clear the window */
   int86(0x10, &inregs, &outregs);     /* execute video interrupt */
   }

dnscroll (ulr, ulc, lrr, lrc, lines)
/*
 *  Scroll a window on the screen down a given number of lines.  If the
 *  number of lines to be scrolled exceeds the number of rows available
 *  in the window, then the area is cleared rather than scrolled.
 *
 *  Parameters:
 *     ulr:    upper, left row of window
 *     ulc:    upper, left column of window
 *     lrr:    lower, right row of window
 *     lrc:    lower, right column of window
 *     lines:  number of lines to scroll window
 *
 *  Return value:
 *     none
 */
int ulr, ulc, lrr, lrc, lines;
   {
   union REGS inregs, outregs;

   inregs.h.ah=0x07;                   /* service code to scroll down */
   inregs.h.bh=0x07;                   /* attribute for new lines */
   inregs.h.ch=--ulr;                  /* upper left row */
   inregs.h.cl=--ulc;                  /* upper left column */
   inregs.h.dh=--lrr;                  /* lower right row */
   inregs.h.dl=--lrc;                  /* lower right column */
   if ((lrr-ulr+1) > lines)            /* see if window is big enough */
      inregs.h.al=lines;               /* number of lines to scroll */
   else
      inregs.h.al=0;                   /* clear the window */
   int86(0x10, &inregs, &outregs);     /* execute video interrupt */
   }

clrwindo (ulr, ulc, lrr, lrc)
/*
 *  Clear a window on the screen.
 *
 *  Parameters:
 *     ulr:    upper, left row of window
 *     ulc:    upper, left column of window
 *     lrr:    lower, right row of window
 *     lrc:    lower, right column of window
 *
 *  Return value:
 *     none
 */
int ulr, ulc, lrr, lrc;
   {
   union REGS inregs, outregs;

   inregs.x.ax=0x0700;                 /* service code to clear window */
   inregs.h.bh=0x07;                   /* attribute for new lines */
   inregs.h.ch=--ulr;                  /* upper left row */
   inregs.h.cl=--ulc;                  /* upper left column */
   inregs.h.dh=--lrr;                  /* lower right row */
   inregs.h.dl=--lrc;                  /* lower right column */
   int86(0x10, &inregs, &outregs);     /* execute video interrupt */
   }

cls ()
/*
 *  Clear the entire screen.
 *
 *  Parameters:
 *     none
 *
 *  Return value:
 *     none
 */
   {
   clrwindo(1, 1, 25, 80);             /* clear the screen */
   locate(1, 1);                       /* put the cursor home */
   }

savescrn (buf)
/*
 *  Save the current display screen in a buffer located in the current
 *  data segment.
 *
 *  Parameters:
 *     buf:  area where screen should be saved
 *
 *  Return value:
 *     none
 */
char *buf;
   {
   int dataseg, dataoff, vidseg;

   #if defined(M_I86SM) || defined(M_I86MM)
      dataseg=(int)getds();            /* use 16-bit data pointer */
      dataoff=(int)buf;
   #else
      dataseg=FP_SEG(buf);             /* use 32-bit data pointer */
      dataoff=FP_OFF(buf);
   #endif

   vidseg=(ismono()) ? 0xb000 : 0xb800;
   movedata(vidseg, 0, dataseg, dataoff, 4096);
   }

rstrscrn (buf)
/*
 *  Restore a screen that has been saved using savescrn().
 *
 *  Parameters:
 *     buf:  area where screen is currently stored
 *
 *  Return value:
 *     none
 */
char *buf;
   {
   int dataseg, dataoff, vidseg;

   #if defined(M_I86SM) || defined(M_I86MM)
      dataseg=(int)getds();            /* use 16-bit data pointer */
      dataoff=(int)buf;
   #else
      dataseg=FP_SEG(buf);             /* use 32-bit data pointer */
      dataoff=FP_OFF(buf);
   #endif

   vidseg=(ismono()) ? 0xb000 : 0xb800;
   movedata(dataseg, dataoff, vidseg, 0, 4096);
   }

savewndo (ulr, ulc, lrr, lrc, bufptr)
/*
 *  Save a window on the screen to some preallocated buffer.
 *
 *  Parameters:
 *     ulr:     upper, left row of window
 *     ulc:     upper, left column of window
 *     lrr:     lower, right row of window
 *     lrc:     lower, right column of window
 *     bufptr:  pointer to area where window is to be stored
 *
 *  Return value:
 *     none
 */
int  ulr, ulc, lrr, lrc;
char *bufptr;
   {
   register int i, vidoff;
   int          dataseg, dataoff, vidseg, rows, cols;

   #if defined(M_I86SM) || defined(M_I86MM)
      dataseg=(int)getds();            /* use 16-bit data pointer */
      dataoff=(int)bufptr;
   #else
      dataseg=FP_SEG(bufptr);          /* use 32-bit data pointer */
      dataoff=FP_OFF(bufptr);
   #endif

   vidseg=(ismono()) ? 0xb000 : 0xb800;/* set the video segment */
   vidoff=scrloc(ulr, ulc);            /* offset into video buffer */
   rows=lrr-ulr+1;                     /* number of rows to save */
   cols=(lrc-ulc+1)*2;                 /* number of columns to save */

   for (i=0; i<rows; i++, vidoff+=160, dataoff+=cols)
      movedata(vidseg, vidoff, dataseg, dataoff, cols);
   }

rstrwndo (ulr, ulc, lrr, lrc, bufptr)
/*
 *  Restore a window to the screen that has been saved using savewndo().
 *
 *  Parameters:
 *     ulr:     upper, left row of window
 *     ulc:     upper, left column of window
 *     lrr:     lower, right row of window
 *     lrc:     lower, right column of window
 *     bufptr:  pointer to area where window is stored
 *
 *  Return value:
 *     none
 */
int  ulr, ulc, lrr, lrc;
char *bufptr;
   {
   register int i, vidoff;
   int          dataseg, dataoff, vidseg, rows, cols;

   #if defined(M_I86SM) || defined(M_I86MM)
      dataseg=(int)getds();            /* use 16-bit data pointer */
      dataoff=(int)bufptr;
   #else
      dataseg=FP_SEG(bufptr);          /* use 32-bit data pointer */
      dataoff=FP_OFF(bufptr);
   #endif

   vidseg=(ismono()) ? 0xb000 : 0xb800;/* set the video segment */
   vidoff=scrloc(ulr, ulc);            /* offset into video buffer */
   rows=lrr-ulr+1;                     /* number of rows to save */
   cols=(lrc-ulc+1)*2;                 /* number of columns to save */

   for (i=0; i<rows; i++, vidoff+=160, dataoff+=cols)
      movedata(dataseg, dataoff, vidseg, vidoff, cols);
   }

exist (file)
/*
 *  Check and see if the specified file exists.
 *
 *  Parameters:
 *     file:  string containing the file/path name to check
 *
 *  Return value:
 *     TRUE if the file was found and FALSE if not
 */
char *file;                            /* file name */
   {
   int rc;

   if ((rc=access(file, 00)) < 0)      /* check file's existence */
      return (FALSE);                  /* file not found */
   else
      return (TRUE);                   /* found it */
   }

char *concat (trim /*, _va */)
/*
 *  Concatenate a number of strings together and place them in a single
 *  array.  The strings may be optionally trimmed of whitespace characters
 *  as they are being concatenated depending on the value of the 'trim'
 *  parameter.  The maximum size of the new string is 200 characters.
 *
 *  Parameters:
 *     trim:  flag indicating if strings should be trimmed (true/false)
 *     _va:   a variable number of string pointers--terminated by a null
 *            pointer--are also provided by the caller
 *
 *  Return value:
 *     a pointer to the new string
 */
int trim;
   {
   va_list       argptr;
   register char *str;
   static char   newstr[200];

   memset(newstr, '\0', 200);          /* init the new string */
   va_start(argptr, trim);             /* point to start of arg list */
   while (str=va_arg(argptr, char *))  /* grab a pointer from the list */
      {
      strcat(newstr, str);             /* append the string */
      if (trim)                        /* trim the string if needed */
         trimall(newstr);
      }
   return (newstr);
   }

unsigned getds ()
/*
 *  Return the value of the current data segment.
 *
 *  Parameters:
 *     none
 *
 *  Return value:
 *     the current data segment
 */
   {
   struct SREGS segregs;

   segread(&segregs);
   return (segregs.ds);
   }
